from __future__ import annotations

from dataclasses import dataclass
from scipp import Variable
from .component import Component


@dataclass
class Attenuator(Component):
    """Only reduces beam intensity without modifying, e.g., divergence

    Use, e.g., Filter with 'AcrylicGlass_C5O2H8' or 'Polycarbonate_C16O3H14'
    NCrystal materials if more-accurate incoherent scattering (from Acrylic or
    Polycarbonate) is desired.
    """
    transmission: Variable


@dataclass
class Filter(Component):
    """A powder or amorphous material that effects a beam with physics from NCrystal

    Likely useful NCrystal data sets include:
        AcrylicGlass_C5O2H8
        Al_sg225
        Be_sg194
        Polycarbonate_C16O3H14
        C_sg194_pyrolytic_graphite
    """
    width: Variable
    height: Variable
    length: Variable
    composition: str
    temperature: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import scalar as s, vector as v
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal['name']
        width = cal.get('width')
        height = cal.get('height')
        length = cal.get('length')
        composition = cal.get('composition')
        temperature = cal.get('temperature', s(300., unit='K'))
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0.,0, 0], unit='deg')))
        return cls(name, position, orientation, width, height, length, composition, temperature)

    def __mccode__(self) -> tuple[str, dict]:
        params = dict()
        params['xwidth'] = self.width.to(unit='m').value
        params['yheight'] = self.height.to(unit='m').value
        params['zdepth'] = self.length.to(unit='m').value
        if '.ncmat' in self.composition:
            sample = self.composition
        else:
            sample = f'{self.composition}.ncmat'
        if ';temp=' not in sample:
            sample = f'{sample};temp={self.temperature:c}'.replace(' ','')
        params['cfg'] = '"' + sample + '"'
        return 'NCrystal_sample', params


@dataclass
class OrderedFilter(Filter):
    """(likely) A Bragg scattering filter, e.g., Pyrolytic Graphite"""
    tau: Variable # the direction and lattice spacing used in the filter


def make_aluminum(name, position, orientation, width, height, length):
    return Filter(name, position, orientation, width, height, length, 'Al_sg225')



from scipp import Variable

from niess.spatial import at_relative_dict


def radius_of_curvature_to_rotation_angle(distance: Variable, radius_of_curvature: Variable) -> Variable:
    """
    Straight segments approximating a curve must be rotated proportional to their length
    """
    from scipp import atan2
    return atan2(y=distance.to(unit=radius_of_curvature.unit)/2, x=radius_of_curvature) * 2


def radius_of_curvature_to_rotation(distance: Variable, radius_of_curvature: Variable) -> Variable:
    """Return the correct quaternion to rotate a BIFROST curved guide element"""
    # TODO verify rotation direction around y
    from scipp import vector
    from scipp.spatial import rotations_from_rotvecs
    theta = radius_of_curvature_to_rotation_angle(distance, radius_of_curvature)
    return rotations_from_rotvecs(theta * vector([0, 1, 0.]))


def half_rotation(rot: Variable) -> Variable:
    from scipp.spatial import rotation
    from scipy.spatial.transform import Rotation
    identity = rotation(value=[0, 0, 0, 1])
    half_rot = rotation(value=Rotation.from_quat(rot.value + identity.value).as_quat())
    return half_rot


def guide_rotated_position(axis: Variable, width: Variable, rot: Variable) -> Variable:
    """Define the position where a rotated guide just touches the corner of the
    previous guide
    :param axis: the symmetry axis of the guide
    :param width: the width of the guide in the direction perpendicular to the
                  guide symmetry axis and the rotation axis
    :param rot: the rotation quaternion of the guide
    """
    from scipp import acos, sin, scalar
    #   *       ..--*     The vector forms a chord of the circle centered on one
    #   |    .-^  .^      edge of the guide, with radius equal to the half of the
    #   | .-^ .--^        width of the guide. It is rotated by the rotation angle
    #   |=--^'            away from the symmetry axis of the guide, such that it has
    #                     distance = 2 * width/2 * sin(angle/2)
    # -> Find the angle of half of the rotation, take its sin, multipy by the width
    # scalar part of quaternion = cos(angle/2)
    distance = width * sin(acos(scalar(half_rotation(rot).value[-1])))
    return distance * (rot * axis)


def radius_of_curvature_to_next_point(point: Variable, distance: Variable, width: Variable, radius_of_curvature: Variable) -> Variable:
    """
    Find the location of this guide relative to the previous guide end

    :param point: the location of the end of the previous guide element
    :param distance: the length of this guide element
    :param width: the width of both guides in the direction perpendicular
    :param radius_of_curvature: the radius of curvature of the whole guide system
    :return: the position of this guide that, when rotated, the two guides touch at
             their inner corner
    """
    from scipp import vector
    axis = vector([0, 0, 1.])
    rotation = radius_of_curvature_to_rotation(distance, radius_of_curvature)
    offset = guide_rotated_position(axis, width, rotation)
    return point + offset


def curved_guide_unit_dictionary(start_at, start_rot, lengths, name, radius_of_curvature, **consts):
    """consts _must_ include 'width' and 'height'"""
    from scipp import vector
    # The local guide direction is _always_ [001]
    z = vector([0, 0, 1.])
    d = {
        f'{name}_guide_{i}': {'length': length, **consts}
        for i, length in enumerate(lengths)
    }
    this = f'{name}_guide_0'
    d[this]['position'] = start_at
    d[this]['orientation'] = start_rot
    end = at_relative_dict(d[this], d[this]['length'] * z)
    end_rot = d[this]['orientation']
    for i, length in enumerate(lengths[1:]):
        prev = this
        this = f'{name}_guide_{i+1}'
        d[this]['position'] = radius_of_curvature_to_next_point(end, length, d[this]['width'], radius_of_curvature)
        d[this]['orientation'] = radius_of_curvature_to_rotation(length, radius_of_curvature) * d[prev]['orientation']
        end = at_relative_dict(d[this], d[this]['length'] * z)
        end_rot = d[this]['orientation']
    return d, end, end_rot


def curved_guide_partial_dict(ref_p, ref_r, radius, table, lengths, min_unit, max_unit, **consts):
    from scipp import vector
    from .guide_tools import Type, parse_guide_table
    from ..spatial import at_relative
    in_section = False
    section = 0
    d = {}
    for entry in parse_guide_table(table):
        if Type.guide == entry[0]:
            in_section = min_unit <= entry[2] <= max_unit
            section = entry[2]
        if in_section:
            if Type.guide == entry[0]:
                ttype, number, identifier, name, length = entry
                fname = f'unit_{section}_{name}'
                d[fname], ref_p, ref_r = curved_guide_unit_dictionary(
                    ref_p,
                    radius_of_curvature_to_rotation(lengths[section][0], radius) * ref_r,
                    lengths[section],
                    fname,
                    radius,
                    **consts
                )
            elif Type.window == entry[0]:
                raise ValueError('No windows in continuous sections!')
            else:
                # a gap, update the position
                ref_p = at_relative(ref_p, ref_r, entry[2] * vector([0, 0, 1.]))
        if Type.guide == entry[0]:
            # After the Unit-8 guide, we are not _in_ the section any longer
            in_section = min_unit <= entry[2] < max_unit

    return d, ref_p, ref_r

def curved_guide_device_partial_dict(ref_p, ref_r, dev_name, dev_dict, table, pre_unit, post_unit, **consts):
    from scipp import vector
    from .guide_tools import Desc, Type, parse_guide_table
    from ..spatial import at_relative
    section = 0
    d = {}
    pre_ok = False
    post_ok = False
    is_exit = True
    for entry in parse_guide_table(table):
        if Type.guide == entry[0]:
            pre_ok = entry[2] >= pre_unit
            post_ok = entry[2] < post_unit
            section = entry[2]
        if pre_ok and post_ok and Type.guide != entry[0]:
            dist = entry[-1]
            if Type.window == entry[0]:
                n, s = (section, 'exit') if is_exit else (section + 1, 'entry')
                d[f'unit_{n}_{s}_window'] = {
                    'position': ref_p,
                    'orientation': ref_r,
                    'length': dist,
                    'composition': 'Al_sg225',
                    **consts
                }
                is_exit = False
            elif Type.gap == entry[0]:
                if Desc.device_gap == entry[1]:
                    # For lack of better information, position the device in the
                    # center of the gap left for it:
                    half_p = at_relative(ref_p, ref_r, dist/2 * vector([0, 0, 1.]))
                    if 'offset' in dev_dict:
                        half_p -= dev_dict['offset']
                    d[dev_name] = {'position': half_p, 'orientation': ref_r}
                    d[dev_name].update(dev_dict)
            else:
                raise ValueError('Only windows or gaps expected here!')
            ref_p = at_relative(ref_p, ref_r, dist * vector([0, 0, 1.]))

    return d, ref_p, ref_r


def curved_guide_parameters(guide_start_vec, guide_start_rot, bunker_chopper_height) -> tuple[dict, Variable, Variable]:
    """
    Parameter dictionary for the post-Pulse Shaping Chopper guides

    As-designed parameters for the BIFROST primary guide system.
    Can and should be replaced by as-built parameters once known.

    The names of dictionary keys here must match attributes of the Primary
    class, the order should not be important.
    """
    from scipp import scalar, vector, array
    from ..spatial import at_relative, at_relative_dict
    m = scalar(1.0, unit='m')
    mm = scalar(1.0, unit='mm').to(unit='m')

    # Described in ESS-1075014, section 3.3

    # Sub-Section 2 is the curved guide after the Pulse Shaping chopper.
    #
    # The order of components is:
    # 1. Segment of guide inside the PSC housing, 'unit 3'
    # 2. PSC guide housing window
    # 3. First beam monitor, a fission chamber that is 15 mm thick.
    #    The gap it sits in may be bigger -- apparently 50 mm, probably including both windows.
    # 4. Aluminum entrance window
    # 5. guide unit 4; segments offset and oriented to approximate a curve
    # 6. First frame-overlap chopper (connected to up-steam and down-stream housing without a window?).
    # 7. guide units 5-8; segments offset and oriented to approximate a curve
    # 8. Second frame-overlap chopper
    # 9. guide units 9-15; segments offset and oriented to approximate a curve

    # Taken from SwissNeutronics drawings for guide sub-assemblies from 2021-07-01
    curved_guide_unit_labels = {
        3: 45752, 4: 45753, 5: 45754, 6: 45755, 7: 45756, 8: 45757, 9: 45758,
        10: 45759, 11: 45760, 12: 45761, 13: 45762, 14: 45763, 15: 45764
    }
    curved_guide_unit_lengths = {
        3: [389.5 * mm],
        4: [500.0 * mm, 500.0 * mm, 500.0 * mm, 177.35 * mm],
        5: [502.78 * mm, 502.78 * mm, 502.78 * mm],
        6: [500.0 * mm, 500.0 * mm, 500.0 * mm, 140 * mm],
        7: [118.3 * mm, 500.0 * mm, 500.0 * mm, 500.0 * mm],
        8: [500.0 * mm, 500.0 * mm, 118.3 * mm, 500.0 * mm],
        9: [500.0 * mm, 500.0 * mm, 500.0 * mm],
        10: [175.71 * mm, 351.42 * mm, 351.42 * mm, 175.71 * mm], # Cu substrate
        11: [236.32 * mm, 472.63 * mm, 236.32 * mm], # Cu substrate
        12: [500.0 * mm, 500.0 * mm, 448.22 * mm],
        13: [500.0 * mm, 500.0 * mm, 448.22 * mm],
        14: [500.0 * mm, 500.0 * mm, 448.22 * mm],
        15: [500.0 * mm, 500.0 * mm, 448.22 * mm],
    }
    swissneutronics_37835_curved_section_table = (
        ('45604- Unit-1 BBG', 482.5 * mm),
        ('gap', 22.05 * mm),
        ('45605- Unit-2 PSC', 389.5 * mm),
        ('gap', 81.71 * mm),  # PSC1 and PSC2
        ('45752- Unit-3 curved', 389.5 * mm),
        ('device gap', 47.0 * mm),  # PSC fission chamber monitor
        ('window', 0.5 * mm),
        ('window gap', 2.5 * mm),
        ('45753- Unit-4 curved', 1677.35 * mm),
        ('window gap', 2.5 * mm),
        ('window', 0.5 * mm),
        ('device gap', 47.0 * mm),  # FOC1
        ('window', 0.5 * mm),
        ('window gap', 2.5 * mm),
        ('45754- Unit-5 curved', 1508.35 * mm),
        ('unit gap', 0.5 * mm),
        ('45755- Unit-6 curved', 1640.0 * mm),
        ('bellow gap', 4.0 * mm),
        ('45756- Unit-7 curved', 1618.3 * mm),
        ('unit gap', 0.5 * mm),
        ('45757- Unit-8 curved', 1618.3 * mm),
        ('window gap', 2.5 * mm),
        ('window', 0.5 * mm),
        ('device gap', 59.0 * mm),  # FOC2 + monitor?
        ('window', 0.5 * mm),
        ('unit gap', 2.5 * mm),
        ('45758- Unit-9 curved', 1568.3 * mm),
        ('unit gap', 0.5 * mm),
        ('45759- Unit-10 curved', 1054.25 * mm),
        ('unit gap', 0.5 * mm),
        ('45760- Unit-11 curved', 945.25 * mm),
        ('bellow gap', 4.0 * mm),
        ('45761- Unit-12 curved', 1448.22 * mm),
        ('unit gap', 0.5 * mm),
        ('45762- Unit-13 curved', 1448.22 * mm),
        ('bellow gap', 4.0 * mm),
        ('45763- Unit-14 curved', 1448.22 * mm),
        ('unit gap', 0.5 * mm),
        ('45764- Unit-15 curved', 1448.22 * mm),
        ('unit gap', 0.5 * mm),
    )

    # Key values, taken from ESS-1075014 Figure 5 Curved section optical summary
    radius_of_curvature = 1518500 * mm  # as designed
    beam_monitor_1_gap = 50 * mm
    beam_monitor_1_gap = swissneutronics_37835_curved_section_table[5][1] # 47 mm

    entry_to_foc1_gap = 2116.7 * mm
    foc1_gap = 65 * mm
    entry_to_foc2_gap = 8559.75 * mm
    foc2_gap = 65 * mm
    entry_to_collimation_section = 10195.5 * mm
    collimation_section = 2000 * mm  # copper-substrate, 300 mm with thicker walls?
    entry_to_exit = 17995.28 * mm
    coatings = {'left': 3.0, 'right': 3.5, 'top': 2.5, 'bottom': 2.5}

    guide_width = 29.53 * mm
    guide_height = 47.51 * mm

    p = dict()
    y = vector([0, 1., 0])
    z = vector([0, 0, 1.])

    p['unit_3_curved'], last_p, last_r = curved_guide_unit_dictionary(
        guide_start_vec, guide_start_rot,
        curved_guide_unit_lengths[3],
        'unit_3_curved',
        radius_of_curvature,
        width=guide_width,
        height=guide_height,
        **coatings,
    )

    window_thickness = 0.5 * mm
    window_gap = 2.5 * mm
    # This window does not appear to be in the SwissNeutronics optics table
    p['psc_exit_window'] = {
        'position': at_relative(last_p, last_r, window_gap * z),
        'orientation': last_r,
        'length': window_thickness,
        'width':  2 * guide_width,
        'height': 2 * guide_height,
        'composition': 'Al_sg225',
    }

    device_gap = swissneutronics_37835_curved_section_table[5][1]
    window_thickness = swissneutronics_37835_curved_section_table[6][1]
    window_gap = swissneutronics_37835_curved_section_table[7][1]

    p['psc_monitor'] = {
        'position': at_relative(last_p, last_r, device_gap/2 * z),
        'orientation': last_r,
        'thickness': window_thickness,
        'width': 2 * guide_width,
        'height': 2 * guide_height,
    }

    p['curved_entrance_window'] = {
        'position': at_relative(last_p, last_r, device_gap * z),
        'orientation': last_r,
        'length': window_thickness,
        'width': guide_width,
        'height': guide_height,
        'composition': 'Al_sg225',
    }
    # StraightGuides following radius of curvature, between PSC and FOC1
    p['unit_4_curved'], last_p, last_r = curved_guide_unit_dictionary(
        at_relative(last_p, last_r, (device_gap + window_thickness + window_gap) * z),
        radius_of_curvature_to_rotation(curved_guide_unit_lengths[4][0], radius_of_curvature) * last_r,
        curved_guide_unit_lengths[4],
        'unit_4_curved',
        radius_of_curvature,
        width=guide_width,
        height=guide_height,
        **coatings,
    )

    # SwissNeutronics table indicates that there is a window exiting Unit 4
    window_gap = swissneutronics_37835_curved_section_table[9][1]
    window_thickness = swissneutronics_37835_curved_section_table[10][1]
    p['unit_4_exit_window'] = {
        'position': at_relative(last_p, last_r, window_gap * z),
        'orientation': last_r,
        'length': window_thickness,
        'width': guide_width,
        'height': guide_height,
        'composition': 'Al_sg225',
    }
    # Move the reference point to the downstream side of the window
    last_p = at_relative_dict(p['unit_4_exit_window'], window_thickness * z)

    # then the first Frame Overlap Chopper
    device_gap = swissneutronics_37835_curved_section_table[11][1]
    radius = 350 * mm
    offset = -(radius - bunker_chopper_height / 2) * y
    p['frame_overlap_chopper_1'] = {
        'position': at_relative(last_p, last_r, device_gap/2 * z) - offset,
        'orientation': last_r,
        'radius': radius,
        'height': bunker_chopper_height,
        'angle': scalar(38.26, unit='deg'),
        'frequency': scalar(14.0, unit='Hz'),
        'phase': scalar(0., unit='deg'),
        'offset': offset,
    }

    # And then a window before Unit 5
    window_thickness = swissneutronics_37835_curved_section_table[12][1]
    window_gap = swissneutronics_37835_curved_section_table[13][1]
    p['unit_5_entry_window'] = {
        'position': at_relative(last_p, last_r, device_gap * z),
        'orientation': last_r,
        'length': window_thickness,
        'width': guide_width,
        'height': guide_height,
        'composition': 'Al_sg225',
    }
    # Move the reference point from the Unit-4 exit window to the start of Unit-5
    last_p = at_relative_dict(p['unit_5_entry_window'], (window_thickness + window_gap) * z)

    # Units between FOC1 and FOC2 are # 5, 6, 7, 8
    section, last_p, last_r = curved_guide_partial_dict(
        last_p, last_r, radius_of_curvature, swissneutronics_37835_curved_section_table,
        curved_guide_unit_lengths,
        5, 8, width=guide_width, height=guide_height, **coatings
    )
    p.update(section)
    # The reference point and reference orientation were last moved
    # to the down-stream end of Unit-8

    foc = {
        'radius': radius,
        'height': bunker_chopper_height,
        'angle': scalar(52.01, unit='deg'),
        'frequency': scalar(14.0, unit='Hz'),
        'phase': scalar(0., unit='deg'),
        'offset': offset,
    }
    section, last_p, last_r = curved_guide_device_partial_dict(
        last_p, last_r, 'frame_overlap_chopper_2', foc,
        swissneutronics_37835_curved_section_table, 8, 9,
        width=guide_width, height=guide_height, **coatings
    )
    p.update(section)
    # reference point moved to after last gap, so position of guide Unit 9

    section, last_p, last_r = curved_guide_partial_dict(
        last_p, last_r, radius_of_curvature, swissneutronics_37835_curved_section_table,
        curved_guide_unit_lengths,
        9, 16, width=guide_width, height=guide_height, **coatings
    )
    p.update(section)
    # reference point and rotation updated to after guide Unit 15 post-unit gap
    # Note that above call specifies max_unit 16 which is not in this table in order
    # to ensure that the 0.5 mm gap is included

    return p, last_p, last_r


from .arm import Arm
from .tank import Tank
from .bifrost import BIFROST
from .channel import Channel
from .triplet import Triplet
from .analyzer import Analyzer
from .primary import Primary

__all__ = [
    Arm,
    Tank,
    BIFROST,
    Channel,
    Triplet,
    Analyzer,
    Primary,
]



from scipp import Variable


def combine_parameters(objs, sample: Variable):
    """To be used internally by Tank and Channel classes

    Parameters
    ----------
    objs:
        An iterable of objects, each which has a `mcstas_parameters` method
        that returns a dictionary of parameters, including at least the keys
        'distance', 'analyzer', 'detector', 'two_theta'
    """
    from numpy import stack
    parameters = [obj.mcstas_parameters(sample) for obj in objs]
    # switch from a list of dictionaries to a dictionary of arrays
    keys = 'distances', 'analyzer', 'detector', 'two_theta'
    return {k: stack([p[k] for p in parameters], axis=0) for k in keys}



from __future__ import annotations

from dataclasses import dataclass
from scipp import Variable

from mccode_antlr.common.parameters import InstrumentParameter
from .component import Component


ESS_SOURCE_DURATION = Variable(values=2.857e-3, unit='s', dims=None)



@dataclass
class Source(Component):
    pass


@dataclass
class ESSource(Source):
    """Representation of the ESS Butterfly source

    https://github.com/mccode-dev/McCode/blob/main/mcstas-comps/sources/ESS_butterfly.comp
    """
    sector: str
    beamline: int
    height: Variable
    cold_frac: float
    focus_distance: Variable | None
    focus_width: Variable | None
    focus_height: Variable | None
    cold_performance: float
    thermal_performance: float
    wavelength_minimum: Variable | InstrumentParameter | None
    wavelength_maximum: Variable | InstrumentParameter | None
    latest_emission_time: Variable | None
    n_pulses: int | None
    accelerator_power: Variable | None

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import vector as v, scalar as s
        from scipp.spatial import rotations_from_rotvecs as r
        name = cal.get('name', 'ESS_source')
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='rad')))

        sector = cal.get('sector', 'W')
        beamline = cal.get('beamline', 4)
        height = cal.get('height', s(3.0, unit='cm'))
        cold_frac = cal.get('cold_fraction', 0.5)
        focus_distance = cal.get('focus_distance', None)
        focus_width = cal.get('focus_width', None)
        focus_height = cal.get('focus_height', None)
        cold_performance = cal.get('cold_performance', 1.0)
        thermal_performance = cal.get('thermal_performance', 1.0)
        wavelength_minimum = cal.get('wavelength_minimum', None)
        wavelength_maximum = cal.get('wavelength_maximum', None)
        if isinstance(wavelength_minimum, str):
            wavelength_minimum = InstrumentParameter.parse(wavelength_minimum)
        if isinstance(wavelength_maximum, str):
            wavelength_maximum = InstrumentParameter.parse(wavelength_maximum)
        latest_emission_time = cal.get('latest_emission_time', None)
        n_pulses = cal.get('n_pulses', None)
        accelerator_power = cal.get('accelerator_power', None)

        return cls(name, position, orientation, sector, beamline, height, cold_frac,
                   focus_distance, focus_width, focus_height, cold_performance,
                   thermal_performance, wavelength_minimum, wavelength_maximum,
                   latest_emission_time, n_pulses, accelerator_power)

    def __mccode__(self) -> tuple[str, dict]:
        from ..utilities import variable_value_or_parameter as value_or
        pars = {
            'sector': '"' + self.sector.strip('"') + '"',
            'beamline': self.beamline,
            'yheight': self.height.to(unit='m').value,
            'cold_frac': self.cold_frac,
            'c_performance': self.cold_performance,
            't_performance': self.thermal_performance,
        }
        if all(x is not None for x in (self.focus_width, self.focus_height, self.focus_distance)):
            pars['dist'] = self.focus_distance.to(unit='m').value
            pars['focus_xw'] = self.focus_width.to(unit='m').value
            pars['focus_yh'] = self.focus_height.to(unit='m').value
        if all(x is not None for x in (self.wavelength_minimum, self.wavelength_maximum)):
            pars['Lmin'] = value_or(self.wavelength_minimum, 'angstrom')
            pars['Lmax'] = value_or(self.wavelength_maximum, 'angstrom')
        if self.latest_emission_time is not None:
            multiplier = self.latest_emission_time.to(unit=ESS_SOURCE_DURATION.unit) / ESS_SOURCE_DURATION
            pars['tmax_multiplier'] = multiplier.value
        if self.n_pulses is not None:
            pars['n_pulses'] = self.n_pulses
        if self.accelerator_power is not None:
            pars['power'] = self.accelerator_power.to(unit='MW').value

        return 'ESS_butterfly', pars



from dataclasses import dataclass


def __is_type__(x, t, name):
    if not isinstance(x, t):
        raise RuntimeError(f"{name} must be a {t}")


@dataclass
class Triplet:
    from mccode_antlr.assembler import Assembler
    from ..components import He3Tube
    from scipp import Variable

    tubes: tuple[He3Tube, He3Tube, He3Tube]
    resistances: Variable

    @staticmethod
    def from_calibration(position: Variable, length: Variable, **params):
        """Take (fitting) calibration data and construct the object used to convert events to (Q,E)"""
        # The current crop of inputs is not sufficient to capture all degrees of freedom, but is a start.
        from scipp import sqrt, dot, vector, scalar, Variable
        from scipp.spatial import rotations
        from ..spatial import is_scipp_vector
        from ..components import He3Tube

        map(lambda x: is_scipp_vector(*x), ((position, 'position'),))
        if position.ndim != 1 or 'tube' not in position.dims or position.sizes['tube'] != 3:
            raise RuntimeError("Expected positions for 3 'tube'")

        ori = params.get('orient', None)
        ori = rotations(values=[[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 1]], dims=['tube']) if ori is None else ori

        pressure = params.get('pressure', scalar(1., unit='atm'))
        radius = params.get('radius', scalar(5., unit='mm'))
        elements = params.get('elements', 10)
        resistivity = params.get('resistivity', scalar(140., unit='Ohm/in').to(unit='Ohm/m'))
        map(lambda x: __is_type__(*x), ((pressure, Variable, 'pressure'), (length, Variable, 'length'),
                                        (radius, Variable, 'radius'), (elements, int, 'elements'),
                                        (resistivity, Variable, 'resistivity')))
        # pack the tube parameters
        pack = elements, radius, pressure

        # ensure that there is one resistivity per tube
        from ..utilities import is_scalar
        from scipp import concat
        if is_scalar(resistivity):
            resistivity = concat((resistivity, resistivity, resistivity), dim='tube')

        # Make the oriented tube-axis vector(s)
        axis = ori * (length.to(unit=position.unit) * vector([0, 1., 0]))  # may be a 0-D or 1-D (tube) vector array
        tube_at = position - 0.5 * axis  # should now be a 1-D (tube) vector array
        tube_to = position + 0.5 * axis  # *ditto*
        tubes = (He3Tube(at, to, rho, *pack) for at, to, rho in zip(tube_at, tube_to, resistivity))

        # Define the ex-Tube resistances
        resistance = params.get('resistance', scalar(2, unit='Ohm'))
        if is_scalar(resistance):
            resistance = concat((resistance, resistance), dim='tube')
        if len(resistance) < 3:
            contact = params.get('contact_resistance', scalar(0, unit='Ohm'))
            resistance = concat((contact, resistance, contact), dim='tube')
        for idx, name in enumerate(('resistance_A', 'resistance_01', 'resistance_12', 'resistance_B')):
            if name in params:
                resistance['tube', idx] = params.get(name)

        return Triplet(tuple(tubes), resistance)

    def triangulate(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [tube.triangulate(unit=unit) for tube in self.tubes]
        return combine_triangulations(vts)

    def extreme_path_corners(self, horizontal, vertical, unit=None):
        from ..spatial import combine_extremes
        vs = [tube.extreme_path_corners(horizontal, vertical, unit=unit) for tube in self.tubes]
        return combine_extremes(vs, horizontal, vertical)

    def horizontal_coverage(self, sample_at, analyzer_at, unit=None):
        """Find and return the horizontal coverage angle of the central tube"""
        from scipp import norm, cross, atan2
        unit = unit or 'radian'
        horizontal = cross(self.tube_com()['tube', 1] - analyzer_at, analyzer_at - sample_at)
        horizontal /= norm(horizontal)
        bbox = self.tubes[1].bounding_box(horizontal, unit=sample_at.unit)
        width = bbox['limits', 1] - bbox['limits', 0]
        distance = norm(self.tube_com()['tube', 1] - analyzer_at) + norm(analyzer_at - sample_at)
        return 2 * atan2(y=width/2, x=distance).to(unit=unit)

    def tube_com(self):
        from scipp import concat
        return concat([(t.at + t.to)/2 for t in self.tubes], 'tube')

    def tube_end(self):
        from scipp import concat
        return concat([(t.to - t.at)/2 for t in self.tubes], 'tube')

    def a_over_a_plus_b_edges(self):
        """Points to convert continuous A/(A+B) to discrete segments per tube"""
        from scipp import concat, scalar, cumsum, max
        tr = [1.0 * t.resistance for t in self.tubes]
        rs = [scalar(0., unit='Ohm'), *[x for a in zip(self.resistances, tr) for x in a], self.resistances[-1]]
        # rs is [0, left_contact, left_tube, left_resistor, center_tube, right_resistor, right_tube, right_contact]
        partial_sums = cumsum(concat(rs, dim='tube'))
        return partial_sums / max(partial_sums)

    def a_minus_b_over_a_plus_b_edges(self):
        """Points to convert continuous (A-B)/(A+B) to discrete segments per tube"""
        from scipp import scalar
        return 2 * self.a_over_a_plus_b_edges() - scalar(1)

    def mcstas_parameters(self) -> dict:
        #TODO make this more accurate -- insert vectors into the instrument defined parameters to use here?
        from scipp import sqrt, dot
        # length vector (from one end of each tube to the other)
        lv = [self.tubes[x].to - self.tubes[x].at for x in range(3)]
        # central vector (the position of the center, relative to defining sample position)
        cv = [(self.tubes[x].to + self.tubes[x].at)/2 for x in range(3)]
        # average length -- they *should* all be identical, but maybe they're not (hence the note above)?
        length = sum([sqrt(dot(x, x)).to(unit='m').value for x in lv]) / 3
        # average radius -- ditto, if they're not identical this is wrong and a vector in the instrument is better
        radius = sum([self.tubes[x].radius.to(unit='m').value for x in range(3)]) / 3
        # The distance between the first and third tube centres plus twice the radius is the assembly width
        width = sqrt(dot(cv[2] - cv[0], cv[2] - cv[0])).to(unit='m').value + 2 * radius
        params = dict(
            charge_a='"event_charge_left"',
            charge_b='"event_charge_right"',
            detection_time='"event_time"',
            tube_index_name='"TUBE"',
            N=3,
            width=width,
            height=length,
            radius=radius,
            wires_in_series=1,
        )
        return params

    def to_mccode(self, assembler: Assembler, relative: str, distance: float, name: str,
                  when: str = None, extend: str = None, add_metadata: bool = False,
                  component: str = None, parameters: dict = None):
        if component is None:
            component = 'Detector_tubes'
        base_parameters = self.mcstas_parameters()
        if parameters is not None:
            base_parameters.update(parameters)
        tubes = assembler.component(name, component, at=((0, 0, distance), relative), parameters=base_parameters)
        tubes.WHEN(when)
        tubes.EXTEND(extend)



from dataclasses import dataclass
from niess.utilities import calibration

from niess import He3Tube

def _tube_at(distance, rotation_y, length, radius, resistivity, elements, pressure):
    from scipp import vector
    from scipp.spatial import rotations_from_rotvecs
    y, z = vector([0, 1, 0]), vector([0, 0, 1])
    r = rotations_from_rotvecs(y * rotation_y)
    pos = r * (z * distance)
    top = pos + length * y / 2
    bottom = pos - length * y / 2
    return He3Tube(at=bottom, to=top, radius=radius, resistivity=resistivity, elements=elements, pressure=pressure)


def _elastic_monitor_from_params(params):
    from scipp import scalar
    # There is a possibility that some or all necessary parameters are missing
    # but there are not always good defaults to provide in all cases. What do we do?
    distance = params.get('sample_elastic_monitor_distance', scalar(0., unit='m'))
    angle = params.get('tank_elastic_monitor_angle', scalar(0., unit='deg'))
    length = params.get('elastic_monitor_length', scalar(0., unit='m'))
    radius = params.get('elastic_monitor_radius', params.get('elastic_monitor_width', scalar(0., unit='m') / 2))
    resistivity = scalar(1., unit='Ohm/cm')
    pressure = params.get('elastic_monitor_pressure', scalar(1., unit='atm'))
    mon = _tube_at(distance, angle, length, radius, resistivity, 1, pressure)
    return mon


@dataclass
class Tank:
    from scipp import Variable
    from .channel import Channel
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.instr import Instance

    channels: tuple[Channel, ...]
    monitor: He3Tube

    @staticmethod
    @calibration
    def from_calibration(params: dict):
        from scipp import array, scalar
        from .channel import Channel

        channel_params = [{'variant': x} for x in ('s', 'm', 'l')]
        channel_params = {i: channel_params[i % 3] for i in range(9)}
        # but this can be overridden by specifying an integer-keyed dictionary with the parameters for each channel
        channel_params = params.get('channel_params', channel_params)
        # The central a4 angle for each channel, relative to the reference tank angle
        angles = params.get('angles',
                            array(values=[-40, -30, -20, -10, 0, 10, 20, 30, 40.], unit='degree', dims=['channel']))

        channels = [Channel.from_calibration(angles[i], **channel_params[i]) for i in range(9)]
        return Tank(tuple(channels), _elastic_monitor_from_params(params))

    @staticmethod
    def unique_from_calibration(**params):
        from scipp import array
        from .channel import Channel
        channel_params = [{'variant': x} for x in ('s', 'm', 'l')]
        channel_params = {i: channel_params[i % 3] for i in range(3)}
        # but this can be overridden by specifying an integer-keyed dictionary with the parameters for each channel
        channel_params = params.get('channel_params', channel_params)
        # The central a4 angle for each channel, relative to the reference tank angle
        angles = params.get('angles',
                            array(values=[-40, -30, -20, -10, 0, 10, 20, 30, 40.], unit='degree', dims=['channel']))

        channels = [Channel.from_calibration(angles[i], **channel_params[i]) for i in range(3)]
        return Tank(tuple(channels), _elastic_monitor_from_params(params))

    def to_secondary(self, **params):
        from scipp import vector
        from ..components import IndirectSecondary

        sample_at = params.get('sample', vector([0, 0, 0.], unit='m'))

        detectors = []
        analyzers = []
        a_per_d = []
        for channel in self.channels:
            for arm in channel.pairs:
                analyzers.append(arm.analyzer.central_blade)
                detectors.extend(arm.detector.tubes)
                a_per_d.extend([len(analyzers) - 1 for _ in arm.detector.tubes])

        from scipp import arange
        nc = len(self.channels)
        np = len(self.channels[0].pairs)
        a = arange(start=0, stop=len(analyzers), dim='n').fold('n', sizes={'channel': nc, 'pair': np})
        d = arange(start=0, stop=len(detectors), dim='n').fold('n', sizes={'channel': nc, 'pair': np, 'tube': 3})

        return IndirectSecondary(detectors, analyzers, a_per_d, sample_at, a, d)

    def triangulate_detectors(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [channel.triangulate_detectors(unit=unit) for channel in self.channels]
        return combine_triangulations(vts)

    def triangulate_analyzers(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [channel.triangulate_analyzers(unit=unit) for channel in self.channels]
        return combine_triangulations(vts)

    def triangulate(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [channel.triangulate(unit=unit) for channel in self.channels]
        return combine_triangulations(vts)

    def mcstas_parameters(self, sample: Variable):
        from numpy import hstack
        from .combine import combine_parameters
        # pull out the list of 'distances', 'analyzer', 'detector', 'two_theta'
        # from each channel, and stack them into a single array per parameters
        parameters = combine_parameters(self.channels, sample)
        parameters['channel'] = hstack([channel.sample_space_angle(sample).value for channel in self.channels])
        return parameters

    def rtp_parameters(self, sample: Variable):
        from scipp import concat
        return [concat(q, dim='channel') for q in zip(*[c.rtp_parameters(sample) for c in self.channels])]

    def to_mccode(self, assembler: Assembler, sample: Instance, settings: dict = None, **kwargs):
        from scipp import vector, concat, max
        from ..mccode import ensure_user_var
        ensure_user_var(assembler, 'int', 'secondary_cassette', 'Secondary spectrometer analyzer cassette index')

        origin = vector([0, 0, 0], unit='m')
        positions = [c.sample_space_angle(origin).to(unit='radian').value for c in self.channels]
        cov_xy = [c.coverage(origin, unit='radian') for c in self.channels]
        cov_x = 2 * max(concat([y for _, y in cov_xy], dim='channel')).value

        slits_name = 'slits'
        declared_positions = f'{slits_name}_positions'
        assembler.declare_array('double', declared_positions, positions, source=__file__, line=173)
        slits = assembler.component(slits_name, 'Slit_radial_multi', at=((0, 0, 0,), sample))
        slits.set_parameters(slit_width=cov_x, offset='slitAngle*DEG2RAD',
                             number=len(self.channels), radius='slitDistance', height=0.2,
                             positions=declared_positions)
        slits.EXTEND("secondary_cassette = (SCATTERED) ? 1 + slit : -1;")

        for index, channel in enumerate(self.channels):
            name = f"channel_{1 + index}"
            when = f"{1 + index} == secondary_cassette"
            channel.to_mccode(assembler, sample, name=name, when=when, settings=settings, **kwargs)




def primary_compressor_parameters(guide_start_vec, guide_start_rot) -> dict:
    """Parameter dictionary for the pre-Pulse Shaping Chopper guides

    As-designed parameters for the BIFROST primary guide system.
    Can and should be replaced by as-built parameters once known.

    The names of dictionary keys here must match attributes of the Primary
    class, the order should not be important.
    """
    from scipp import scalar, array, sum, vector
    from ..spatial import at_relative

    mm = scalar(1.0, unit='mm').to(unit='m')
    z = vector([0, 0, 1.])

    # verified guide dimensions reported as part of ESS-4019844
    # coating m-values from same, which also has measured reflectivity curves
    bbg_45604 = {
        'length': 482.5 * mm,
        'left': 1.5, # sn_45607_1
        'right': 1.5 , # sn_45607_2
        'top': 2.5, # sn_45606_1
        'bottom': 2.5, # sn_45606_2
        'entry': {
            'inner': {'width': 50.84 * mm, 'height': 48.8 * mm},
            'outer': {'width': 66.84 * mm, 'height': 64.8 * mm}
        },
        'exit': {
            'inner': {'width': 41.78 * mm, 'height': 48.26 * mm},
            'outer': {'width': 57.78 * mm, 'height': 64.26 * mm}
        },
        'horizontal': {
            'major': 4134.2425 * mm, 'minor': 34.817 * mm, 'offset': -2825.03 * mm
        },
        'vertical': {
            'major': 4621.6055 * mm, 'minor': 24.431 * mm, 'offset': -239.795 * mm
        }
    }
    psc_45605 = {
        'length': 389.5 * mm,
        'left': 1.5, # sn_45609_2
        'right': 1.5, # sn_45609_1
        'top': 2.0, # sn_45608_1
        'bottom': 2.0, # sn_45608_2
        'entry': {
            'inner': {'width': 41.28 * mm, 'height': 48.22 * mm},
            'outer': {'width': 57.28 * mm, 'height': 64.22 * mm}
        },
        'exit': {
            'inner': {'width': 30.41 * mm, 'height': 47.37 * mm},
            'outer': {'width': 46.41 * mm, 'height': 63.37 * mm}
        },
        'horizontal': {
            'major': 4134.2425 * mm, 'minor': 34.817 * mm, 'offset': -3329.584 * mm
        },
        'vertical': {
            'major': 4621.6055 * mm, 'minor': 24.431 * mm, 'offset': -744.345 * mm
        }
    }

    p = dict()

    """
    The first part of 'element 6' of the BIFROST primary spectrometer

    The so-called Neutron Beam Optical Assembly, which is a metal-substrate guide
    inserted into the target monolith.
    The guide itself has 1.5 mm aluminum windows at each end, and the monolith
    has a 4 mm window separating the NBOA from the BBG
    """
    nboa_window_thickness = 1.5 * mm
    nboa_input_height = 32.37 * mm
    nboa_input_width = 44.72 * mm
    nboa_substrate = 12.0 * mm
    nboa_output_height = 50.07 * mm  # might be 50.1
    nboa_output_width = 35.32 * mm

    p['nboa_entry_window'] = {
        'position': at_relative(guide_start_vec, guide_start_rot, -nboa_window_thickness * z),
        'orientation': guide_start_rot,
        'length': nboa_window_thickness,
        'composition': 'Al_sg225',
        'width': nboa_input_width + 4 * nboa_substrate,
        'height': nboa_input_height + 4 * nboa_substrate,
    }

    nboa_segments = array(values=[0.48844444, 0.48844444, 0.48844444, 0.48844444,
                                  0.48844444, 0.48844444, 0.48844444, 0.06121889],
                          dims=['segments'], unit='m')
    nboa_bbg_gap = scalar(0.2349, unit='m')
    bbg_nose_gap = scalar(0.015, unit='m')
    # NBOA parameters from original file included the BBG and PSC Nose lengths
    # plus gaps (including windows) -- account for the extra 'missing' lengths here:
    nboa_beyond_out = nboa_bbg_gap + bbg_nose_gap + bbg_45604['length'] + psc_45605['length']
    p['nboa'] = {
        'position': guide_start_vec,
        'orientation': guide_start_rot,
        'length': nboa_segments,
        'left': (3.0, 3.0, 2.5, 2.5, 2.5, 2.0, 1.5, 1.5),
        'right': (3.0, 3.0, 2.5, 2.5, 2.5, 2.0, 1.5, 1.5),
        'top': (3.5, 3.5, 3.0, 3.0, 2.5, 2.5, 2.5, 2.5),
        'bottom': (3.5, 3.5, 3.0, 3.0, 2.5, 2.5, 2.5, 2.5),
        'horizontal': {
            'midpoint': scalar(0.069634, unit='m'),
            'in': scalar(3.4578, unit='m'),
            'out': scalar(0.415155, unit='m') + nboa_beyond_out,
        },
        'vertical': {
            'midpoint': scalar(0.04862, unit='m'),
            'in': scalar(1.36, unit='m'),
            'out': scalar(3.487681, unit='m') + nboa_beyond_out,
        }
    }
    nboa_length = sum(nboa_segments)

    p['nboa_exit_window'] = {
        'position': at_relative(guide_start_vec, guide_start_rot,
                                (nboa_length + nboa_window_thickness) * z),
        'orientation': guide_start_rot,
        'length': nboa_window_thickness,
        'composition': 'Al_sg225',
        'width': nboa_output_width + 4 * nboa_substrate,
        'height': nboa_output_height + 4 * nboa_substrate,
    }

    p['monolith_window'] = {
        'position': at_relative(p['nboa_exit_window']['position'], guide_start_rot,
                                (10 * mm) * z),
        'orientation': guide_start_rot,
        'length': 4 * mm,
        'composition': 'Al_sg225',
        'width': nboa_input_width + 4 * nboa_substrate,
        'height': nboa_input_height + 4 * nboa_substrate,
    }

    """
    The second part of 'element 6' of the BIFROST primary spectrometer

    Part of the shutter assembly just outside the target monolith. When the shutter
    is closed this guide segment is translated out of the beam path and is replaced by
    neutron blocking material.
    This movable section has a window before and after the guide segment.
    """
    p['bbg_entry_window'] = {
        'position': at_relative(p['nboa_exit_window']['position'], guide_start_rot,
                                (nboa_bbg_gap - 2 * mm) * z),
        'orientation': guide_start_rot,
        'length': 0.5 * mm,
        'composition': 'Al_sg225',
        'width': bbg_45604['entry']['outer']['width'],
        'height': bbg_45604['entry']['outer']['height'],
    }

    # The bridge beam guide is a moveable element which exchanges with a shutter
    p['bbg'] = {
        'position':  at_relative(guide_start_vec, guide_start_rot, (sum(nboa_segments) + nboa_bbg_gap) * z),
        'orientation': guide_start_rot,
    }
    p['bbg'].update(bbg_45604)

    p['bbg_exit_window'] = {
        'position': at_relative(p['bbg']['position'], guide_start_rot,
                                (p['bbg']['length'] + 1 * mm) * z),
        'orientation': guide_start_rot,
        'length': 0.5 * mm,
        'composition': 'Al_sg225',
        'width': bbg_45604['exit']['outer']['width'],
        'height': bbg_45604['exit']['outer']['height'],
    }

    """
    The final part of 'element 6' of the BIFROST primary spectrometer

    A segment of focusing guide inside the vacuum housing of the pulse shaping chopper.
    There is an aluminum window upstream into the vacuum housing, but none between
    this guide and the chopper.
    """
    p['psc_housing_entry_window'] = {
        'position': at_relative(p['bbg_exit_window']['position'], guide_start_rot,
                                (bbg_nose_gap - 2 * mm)  * z),
        'orientation': guide_start_rot,
        'length': 0.5 * mm,
        'composition': 'Al_sg225',
        'width': psc_45605['entry']['outer']['width'],
        'height': psc_45605['entry']['outer']['height'],
    }

    # The last guide segment before the PSC, in its vacuum housing
    p['nose'] = {
        'position': at_relative(p['bbg']['position'], guide_start_rot,
                                (p['bbg']['length'] + bbg_nose_gap) * z),
        'orientation': guide_start_rot,
    }
    p['nose'].update(psc_45605)

    # stash the end of this guide element as a reference point for the first PS chopper
    p['nose']['end'] = at_relative(p['nose']['position'], guide_start_rot, p['nose']['length'] * z)
    return p



from scipp import Variable
from scipp import scalar
mm = scalar(1.0, unit='mm').to(unit='m')

"""
The expanding guide section is made up of two copper-substrate guide units which
pass through the Bunker Wall, then eleven glass substrate guide units of possibly
varying material.

The copper guides are (probably) elliptical in construction.
The glass guide are (probably) not elliptical, but instead made of tapered segments
which closely approximate the elliptical geometry.

For now all guide units use the perfect elliptical approximation for expediency.
If necessary, the individual guide segments lengths of each unit are recorded to allow
using the tapered approximation.
"""


def guide_segment_lengths():
    return {
        45717: [249.75 * mm , 500 * mm, 500 * mm, 500 * mm],
        45718: [249.75 * mm , 500 * mm, 500 * mm, 500 * mm],
        45765: [500 * mm , 500 * mm, 500 * mm, 445.1 * mm],
        45766: [500 * mm , 500 * mm, 500 * mm, 445.1 * mm],
        45767: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45768: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45769: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45770: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45771: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45772: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45773: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45774: [499.94 * mm, 499.94 * mm, 499.94 * mm, 499.94 * mm],
        45775: [500 * mm, 500 * mm, 465.28 * mm]
    }


def guide_unit_parameters():
    """
    Guide parameters taken from the SwissNeutronics drawings from 13 July 2021.
    """
    # Since the ellipse axes are constant through this section, we can add them
    # at the end, so this first dictionary tracks only the unique data per drawing:
    #   {unit number: {length, offset, horizontal m value, vertical m value}}
    values = {
        45717: {'l': 1749.75 * mm, 'o': -24928.98 * mm, 'h': 2.5, 'v': 2.0},
        45718: {'l': 1749.75 * mm, 'o': -23178.73 * mm, 'h': 2.5, 'v': 2.0},
        45765: {'l': 1945.10 * mm, 'o': -21375.98 * mm, 'h': 2.0, 'v': 2.0},
        45766: {'l': 1945.10 * mm, 'o': -19430.38 * mm, 'h': 2.0, 'v': 2.0},
        45767: {'l': 1999.75 * mm, 'o': -17481.28 * mm, 'h': 2.0, 'v': 1.5},
        45768: {'l': 1999.75 * mm, 'o': -15481.03 * mm, 'h': 2.0, 'v': 1.5},
        45769: {'l': 1999.75 * mm, 'o': -13477.28 * mm, 'h': 1.5, 'v': 1.5},
        45770: {'l': 1999.75 * mm, 'o': -11477.03 * mm, 'h': 1.5, 'v': 1.5},
        45771: {'l': 1999.75 * mm, 'o': -9473.28 * mm, 'h': 1.5, 'v': 1.5},
        45772: {'l': 1999.75 * mm, 'o': -7473.03 * mm, 'h': 1.5, 'v': 1.5},
        45773: {'l': 1999.75 * mm, 'o': -5469.28 * mm, 'h': 1.5, 'v': 1.5},
        45774: {'l': 1999.75 * mm, 'o': -3469.03 * mm, 'h': 1.5, 'v': 1.5},
        45775: {'l': 1465.28 * mm, 'o': -1465.28 * mm, 'h': 1.5, 'v': 1.5},
    }
    pars = {
        k: {
            'length': v['l'],
            'vertical': {
                'major': 29352.6625 * mm, 'minor': 45 * mm, 'offset': v['o']
            },
            'horizontal': {
                'major': 28638.5425 * mm, 'minor': 30 * mm, 'offset': v['o']
            },
            'left': v['h'],
            'right': v['h'],
            'top': v['v'],
            'bottom': v['v'],
        } for k, v in values.items()
    }
    return pars


def guide_table():
    from .guide_tools import parse_guide_table
    swissneutronics_37835_expanding_section_table = (
        ('45717- Unit-16 BW insert', 1749.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45718- Unit-17 BW insert', 1749.75 * mm),
        ('window gap', 2.0 * mm),
        ('window', 0.5 * mm),
        ('device gap', 44.5 * mm),
        ('window', 0.5 * mm),
        ('window gap', 2.5 * mm),
        ('boral mask', 3.0 * mm),
        ('45765- Unit-18 expanding', 1945.1 * mm),
        ('unit gap', 0.5 * mm),
        ('45766- Unit-19 expanding', 1945.1 * mm),
        ('bellow gap', 4.0 * mm),
        ('45767- Unit-20 expanding', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45768- Unit-21 expanding', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45769- Unit-22 expanding', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45770- Unit-23 expanding', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45771- Unit-24 expanding', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45772- Unit-25 expanding', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45773- Unit-26 expanding', 1999.75 * mm),
        ('unit gap', 0.5 * mm),
        ('45774- Unit-27 expanding', 1999.75 * mm),
        ('bellow gap', 4.0 * mm),
        ('45775- Unit-28 expanding', 1465.28 * mm),
        ('window gap', 2.5 * mm),
        ('window', 0.5 * mm),
        ('device gap', 114.0 * mm),
    )
    return parse_guide_table(swissneutronics_37835_expanding_section_table)


def unit_dict(ref_p, ref_r, number, length):
    from .guide_tools import straight_unit_dict
    params = guide_unit_parameters()
    return straight_unit_dict(ref_p, ref_r, params[number])


def expanding_guide_parameters(guide_pos, guide_rot) -> tuple[dict, Variable, Variable]:
    # Component order for expanding guide:
    # 1. Copper substrate elliptical guide, Bunker Wall Insert; units 16 & 17
    # 2. Beam monitor 2
    # 3. Glass substrate elliptical-approximation tapered guide segments; units 18-28
    # 4. Shutter to allow entry into the cave without closing the BBG shutter
    from .guide_tools import guide_partial_dict, device_partial_dict, exiting_partial_dict

    table = guide_table()

    p = {}
    bwi, ref_p, ref_r = guide_partial_dict(guide_pos, guide_rot, table, 16, 17, unit_dict)
    p.update(bwi)

    sizes = {'width': 52 * mm, 'height': 72 * mm,}
    device = ('overlap_monitor', {
        'thickness': 0.1 * mm,  # I-BM 100 has a 100 micron thick foil ... plus other stuff in the beam
        'sample_rate': scalar(70., unit='kHz'),
        **sizes,
    })
    d, ref_p, ref_r = device_partial_dict(ref_p, ref_r, (device,), table, 17, 18, sizes)
    p.update(d)

    d, ref_p, ref_r = guide_partial_dict(ref_p, ref_r, table, 18, 28, unit_dict)
    p.update(d)

    beam = {'width': 60 * mm, 'height': 90 * mm}
    window = {k: v + 20 * mm for k, v in beam.items()}
    # there is a shutter in this gap, but there's no need to include it for McStas
    # d, ref_p, ref_r = device_partial_dict(ref_p, ref_r, None, table, 28, 29, window)
    # ignoring that there _is_ a device to be inserted causes a logic error
    d, ref_p, ref_r = exiting_partial_dict(ref_p, ref_r, table, window)
    p.update(d)


    return p, ref_p, ref_r



from dataclasses import dataclass


def variant_parameters(params: dict, default: dict):
    variant = params.get('variant', default['variant'])
    complete = {k: params.get(k, v[variant] if isinstance(v, dict) else v) for k, v in default.items()}
    return complete


@dataclass
class Channel:
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.instr import Instance
    from scipp import Variable
    from .arm import Arm

    pairs: tuple[Arm, Arm, Arm, Arm, Arm]

    @staticmethod
    def from_calibration(relative_angle: Variable, **params):
        from math import pi
        from scipp import sqrt, tan, atan, asin, min, vector
        from scipp.constants import hbar, neutron_mass
        from scipp.spatial import rotations_from_rotvecs
        from .parameters import known_channel_params, tube_xz_displacement_to_quaternion
        from .arm import Arm

        vp = variant_parameters(params, known_channel_params())

        tau = params.get('tau', 2 * pi / vp['d_spacing'])

        sample = vp['sample']

        analyzer_vector = vector([1, 0, 0]) * vp['sample_analyzer_distance']

        ks = (sqrt(vp['energy'] * 2 * neutron_mass) / hbar).to(unit='1/angstrom')
        two_thetas = -2 * asin(0.5 * tau / ks)
        two_theta_vectors = two_thetas * vector([0, -1, 0])
        two_theta_rotation = rotations_from_rotvecs(two_theta_vectors)

        # Detector offsets are specified in a frame with x along the scattered beam, y in the plane of the analyzer
        add = 'analyzer_detector_distance'
        detector_vector = vector([1, 0, 0]) * vp[add] + vp['detector_offset'].to(unit=vp[add].unit)

        # Rotation of the whole analyzer channel around the vertical sample-table axis
        relative_rotation = rotations_from_rotvecs(relative_angle * vector([0, 0, 1]))

        analyzer_position = sample + relative_rotation * analyzer_vector
        detector_position = sample + relative_rotation * (analyzer_vector + two_theta_rotation * detector_vector)

        tau_vecs = relative_rotation * rotations_from_rotvecs(0.5 * two_theta_vectors) * (tau * vector([0, 0, -1]))

        # The detector orientation is given by a displacement vector of the tube-end, we want the associated quaternion
        detector_orient = tube_xz_displacement_to_quaternion(vp['detector_length'], vp['detector_orient'])

        # The detector tube orientation rotation(s) must be modified by the channel rotation:
        detector_orient = relative_rotation * detector_orient

        # vp['coverage'] is specified at 2.7 meV as +/- 2 degrees
        # and higher energies keep constant delta-Q-out-of-plane
        # this coverage is still only half, since both +/- directions have same Q extent
        coverages = atan(min(ks) * tan(1.0 * vp['coverage']) / ks)

        # print(f"Vertical coverage = {2*coverages.to(unit='degree'):c}")
        mosaic = vp['crystal_mosaic']

        resistance = vp['resistance']
        resistivity = vp['resistivity']
        from ..utilities import is_scalar
        from scipp import concat
        if is_scalar(resistance):
            contact_resistance = vp['contact_resistance']
            resistance = concat((contact_resistance, resistance, resistance, contact_resistance), dim='tube')
        if is_scalar(resistivity):
            resistivity = concat((resistivity, resistivity, resistivity), dim='tube')

        def idx_or(obj, index):
            return obj['analyzer', index] if 'analyzer' in obj.dims else obj

        pairs = []
        for idx, (ap, tv, dl, ct, cs, cc, gp) in enumerate(zip(
                analyzer_position, tau_vecs, vp['detector_length'], vp['blade_count'], vp['crystal_shape'], coverages,
                vp['gap']
        )):
            params = dict(
                sample=sample,
                blade_count=ct,
                shape=cs,
                mosaic=idx_or(mosaic, idx),
                analyzer_orient=relative_rotation,
                coverage=cc,
                detector_orient=idx_or(detector_orient, idx),
                resistance=idx_or(resistance, idx),
                resistivity=idx_or(resistivity, idx),
                gap=gp
            )
            pairs.append(Arm.from_calibration(ap, tv, detector_position['analyzer', idx], dl, **params))

        return Channel((pairs[0], pairs[1], pairs[2], pairs[3], pairs[4]))

    def triangulate_detectors(self, unit=None):
        from ..spatial import combine_triangulations
        return combine_triangulations([arm.triangulate_detector(unit=unit) for arm in self.pairs])

    def triangulate_analyzers(self, unit=None):
        from ..spatial import combine_triangulations
        return combine_triangulations([arm.triangulate_analyzer(unit=unit) for arm in self.pairs])

    def triangulate(self, unit=None):
        from ..spatial import combine_triangulations
        return combine_triangulations([arm.triangulate(unit=unit) for arm in self.pairs])

    def mcstas_parameters(self, sample: Variable):
        from .combine import combine_parameters
        return combine_parameters(self.pairs, sample)

    def sample_space_angle(self, sample: Variable):
        return self.pairs[0].sample_space_angle(sample)

    def coverage(self, sample: Variable, unit=None):
        from scipp import concat, max
        unit = unit or 'radian'
        cov_xy = [x.coverage(sample, unit=unit) for x in self.pairs]
        cov_x = max(concat([x for x, _ in cov_xy], dim='pairs'))
        cov_y = max(concat([y for _, y in cov_xy], dim='pairs'))
        return cov_x, cov_y

    def rtp_parameters(self, sample: Variable):
        from scipp import concat, all, isclose
        sa, ad, x, y, angle = zip(*[p.rtp_parameters(sample) for p in self.pairs])
        sa = concat(sa, dim='pairs')
        ad = concat(ad, dim='pairs')
        x7, y7, a7 = [concat(q[:2], dim='pairs') for q in (x, y, angle)]
        x9, y9, a9 = [concat(q[2:], dim='pairs') for q in (x, y, angle)]

        relative_angles = [arm.sample_space_angle(sample) for arm in self.pairs]
        ra0 = relative_angles[0]
        if not all(isclose(concat(relative_angles, dim='arm'), ra0)):
            raise RuntimeError("different relative angles for same-channel analyzers?!")

        return sa, ad, x7, y7, a7, x9, y9, a9, ra0

    def to_mccode(self, assembler: Assembler, relative: Instance, name: str, when: str = None, settings: dict = None, **kwargs):
        from scipp import concat, all, isclose, vector
        # For each channel we need to define the local coordinate system, relative to the provided sample
        origin = vector([0, 0, 0], unit='m')
        ra0 = self.sample_space_angle(origin).to(unit='degree').value
        cassette = assembler.component(f"{name}_arm", "Arm", at=((0, 0, 0), relative), rotate=((0, ra0, 0), relative))
        cassette.WHEN(when)

        for uv in ('int secondary_scattered;', 'int analyzer;', 'int flag;'):
            assembler.ensure_user_var(uv)

        for arm_index, arm in enumerate(self.pairs):
            arm_name = f"{name}_{1 + arm_index}"
            arm_when = f"0 == secondary_scattered && {when}"
            extend = f"secondary_scattered = (SCATTERED) ? 1 : 0;\nanalyzer = (SCATTERED) ? {1 + arm_index} : 0;"
            detector_when = f"{when} && {1 + arm_index}==analyzer"
            detector_extend = f"flag = (SCATTERED) ? 1 : 0;"
            arm.to_mccode(assembler, cassette, name=arm_name, analyzer_when=arm_when, analyzer_extend=extend,
                          settings=settings, detector_when=detector_when, detector_extend=detector_extend, **kwargs)



from dataclasses import dataclass
from scipp import Variable
from mccode_antlr.assembler import Assembler
from .component import Component

@dataclass
class Aperture(Component):
    width: Variable
    height: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        return cls(name, position, orientation, width, height)


@dataclass
class Jaw(Aperture):
    """A special variable width aperture, open by default and configured at runtime"""

    def __mccode__(self) -> tuple[str, dict]:
        params = {
            'xmin': f'{self.name}_l',
            'xmax': f'{self.name}_r',
            'yheight': self.height.to(unit='m').value,
        }
        return 'Slit', params

    def to_mccode(self, assembler: Assembler):
        from ..mccode import ensure_runtime_line as ensure
        half = self.width.to(unit='m').value / 2
        ensure(assembler, f'{self.name}_l/"m" = {-half}')
        ensure(assembler, f'{self.name}_r/"m" = {half}')
        return super().to_mccode(assembler)


@dataclass
class Slit(Aperture):
    """A special variable aperture, open by default and configured at runtime"""

    def __mccode__(self) -> tuple[str, dict]:
        params = {
            'xmin': f'{self.name}_l',
            'xmax': f'{self.name}_r',
            'ymin': f'{self.name}_b',
            'ymax': f'{self.name}_t',
        }
        return 'Slit', params

    def to_mccode(self, assembler: Assembler):
        from ..mccode import ensure_runtime_line as ensure
        half = self.width.to(unit='m').value / 2
        ensure(assembler, f'{self.name}_l/"m" = {-half}')
        ensure(assembler, f'{self.name}_r/"m" = {half}')
        half = self.height.to(unit='m').value / 2
        ensure(assembler, f'{self.name}_b/"m" = {-half}')
        ensure(assembler, f'{self.name}_t/"m" = {half}')
        return super().to_mccode(assembler)



from dataclasses import dataclass, fields
from ..utilities import calibration

# TODO Consider whether it would be possible to use any one of:
#      datclass-wizzard: https://dataclass-wizard.readthedocs.io
#      pydantic: https://docs.pydantic.dev
#      dacite: https://github.com/konradhalas/dacite
#      to handle (de)serializing these nested dataclass objects from calibration data.

@dataclass
class Section:

    @classmethod
    def parts(cls):
        """Get the ordered list of components which make up this Section"""
        # Note to self, one _could_ hard code this instead if the
        # dataclasses.fields(cls) trick stops working at some point
        # but the _order_ of the tuple returned by dataclasses.fields(Type) is
        # (currently) guaranteed to be the same as the order of definition above
        return [field.name for field in fields(cls)]

    @classmethod
    def types(cls):
        """Get the ordered list of component types which make up this Section"""
        return [field.type for field in fields(cls)]

    @classmethod
    def items(cls):
        """Get the ordered list of component names and types which make up this Section"""
        return [(field.name, field.type) for field in fields(cls)]

    def to_mccode(self, *args, **kwargs):
        for part in self.parts():
            getattr(self, part).to_mccode(*args, **kwargs)

    @classmethod
    @calibration
    def from_calibration(cls, parameters: dict):
        for part in cls.parts():
            assert part in parameters

        def named_par(name):
            if 'name' not in parameters[name]:
                parameters[name]['name'] = name
            return parameters[name]

        values = [T.from_calibration(named_par(n)) for n, T in cls.items()]
        return cls(*values)


from dataclasses import dataclass

from mccode_antlr.assembler import Assembler
from scipp import Variable
from .component import Component

# Monitors used on BIFROST, which are limited within McStas to always produce histograms

def nexus_structure_metadata(topic, *, variables, constants):
    from json import dumps
    from mccode_antlr.common import MetaData
    from mccode_to_kafka.writer import da00_dataarray_config
    # TODO uodate eniius.utils.mccode_component_eniius_data to use ('*', 'application/json') instead of
    #      only ('eniius_data', 'json')
    # The correct JSON to encode is a dictionary with a single key, 'data', which itself is a dictionary with
    # the keys 'type' and 'value'. The value of 'type' is 'dict' and the value of 'value' is the actual NeXus structure.
    # This is very hacky and is (currently) necessary to escape parsing within eniius.

    struct = da00_dataarray_config(topic, source='mccode-to-kafka', variables=variables, constants=constants)

    return MetaData.from_instance_tokens(topic, 'application/json', 'nexus_structure_stream_data', dumps(struct))


def add_frame_monitor_metadata(inst, n):
    from mccode_to_kafka.writer import da00_variable_config
    axes = {
        'signal': {'unit': 'counts', 'label': f'{inst.name} counts', 'shape': [n]},
        'errors': {'unit': 'counts', 'label': f'{inst.name} count errors',
                   'shape': [n]},
        't': {'unit': 'microsecond', 'label': 'time since reference', 'shape': [n + 1]},
    }
    configs = {
        k: da00_variable_config(**v, name=k, axes=['t'], data_type='float64')
        for k, v in axes.items()
    }
    variables = [configs['signal'], configs['errors']]
    constants = [configs['t']]
    metadata = nexus_structure_metadata(topic=inst.name, variables=variables,
                                        constants=constants)
    inst.add_metadata(metadata)
    return inst



@dataclass
class FissionChamber(Component):
    """Zero-dimensional fission chamber monitor.
    Outputs events without any spatial information.
    """
    width: Variable
    height: Variable
    thickness: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        thickness = cal.get('thickness', cal.get('length'))
        return cls(name, position, orientation, width, height, thickness)

    def __mccode__(self) -> tuple[str, dict]:
        p = {
            'xwidth': self.width.to(unit='m').value,
            'yheight': self.height.to(unit='m').value,
            'nt': int(1e6 / 14.0 / 7),  # 7 microsecond bins
            'frequency': 14.0,
        }
        return 'Frame_monitor', p

    def to_mccode(self, assembler: Assembler):
        inst = super().to_mccode(assembler)
        # Build the NeXus Structure entry to point to the correct Kafka stream
        return add_frame_monitor_metadata(inst, self.__mccode__()[1]['nt'])


@dataclass
class BeamCurrentMonitor(Component):
    """Zero-dimensional beam current monitor.
    Outputs a current sampled at a configurable frequency.
    """
    width: Variable
    height: Variable
    thickness: Variable
    sample_rate: Variable

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        thickness = cal.get('thickness', cal.get('length'))
        sample_rate = cal.get('sample_rate', cal.get('frequency'))
        if sample_rate is None:
            raise ValueError(f'The sample rate for {name} must be defined')
        return cls(name, position, orientation, width, height, thickness, sample_rate)

    def __mccode__(self) -> tuple[str, dict]:
        from scipp import scalar
        source_frequency = scalar(14.0, unit='Hz')
        nt = int((self.sample_rate.to(unit='Hz') / source_frequency).value)
        p = {
            'xwidth': self.width.to(unit='m').value,
            'yheight': self.height.to(unit='m').value,
            'nt': nt,
            'frequency': source_frequency.to(unit='Hz').value,
        }
        return 'Frame_monitor', p

    def to_mccode(self, assembler: Assembler):
        inst = super().to_mccode(assembler)
        # Build the NeXus Structure entry to point to the correct Kafka stream
        return add_frame_monitor_metadata(inst, self.__mccode__()[1]['nt'])


@dataclass
class GEM2D(Component):
    """Two-dimensional Gas Electron Multiplier monitor.
    Outputs events on X or Y strips (without coincidence) or at (X, Y) point
    (with coincidence).

    For now the pixelation is ignored and a 0-D no-coincidence frame monitor is output
    """
    width: Variable
    height: Variable
    thickness: Variable
    x_strips: int
    y_strips: int

    @classmethod
    def from_calibration(cls, cal: dict):
        name = cal['name']
        position = cal['position']
        orientation = cal['orientation']
        width = cal['width']
        height = cal['height']
        thickness = cal.get('thickness', cal.get('length'))
        x_strips = cal.get('x_strips', cal.get('nx', 1))
        y_strips = cal.get('y_strips', cal.get('ny', 1))
        return cls(name, position, orientation, width, height, thickness, x_strips, y_strips)

    def __mccode__(self) -> tuple[str, dict]:
        from scipp import scalar
        source_frequency = scalar(14.0, unit='Hz')
        nt = 1000 # 71.42 µs bins
        p = {
            'xwidth': self.width.to(unit='m').value,
            'yheight': self.height.to(unit='m').value,
            'nt': nt,
            'frequency': source_frequency.to(unit='Hz').value,
        }
        return 'Frame_monitor', p

    def to_mccode(self, assembler: Assembler):
        inst = super().to_mccode(assembler)
        # Build the NeXus Structure entry to point to the correct Kafka stream
        return add_frame_monitor_metadata(inst, self.__mccode__()[1]['nt'])



# SPDX-FileCopyrightText: 2025-present Gregory Tucker <gregory.tucker@ess.eu>
#
# SPDX-License-Identifier: MIT
from niess.bifrost.primary import Primary


def test_primary_parameters():
    from textwrap import dedent
    from niess.bifrost.parameters import primary_parameters
    parameters = primary_parameters()

    names = list(parameters.keys())
    expected = dedent("""\
    source
    nboa_entry_window
    nboa
    nboa_exit_window
    monolith_window
    bbg_entry_window
    bbg
    bbg_exit_window
    psc_housing_entry_window
    nose
    pulse_shaping_chopper_1
    pulse_shaping_chopper_2
    unit_3_curved
    psc_exit_window
    psc_monitor
    curved_entrance_window
    unit_4_curved
    unit_4_exit_window
    frame_overlap_chopper_1
    unit_5_entry_window
    unit_5_curved
    unit_6_curved
    unit_7_curved
    unit_8_curved
    unit_8_exit_window
    frame_overlap_chopper_2
    unit_9_entry_window
    unit_9_curved
    unit_10_curved
    unit_11_curved
    unit_12_curved
    unit_13_curved
    unit_14_curved
    unit_15_curved
    unit_16_bw_insert
    unit_17_bw_insert
    unit_17_exit_window
    overlap_monitor
    unit_18_entry_window
    unit_18_expanding
    unit_19_expanding
    unit_20_expanding
    unit_21_expanding
    unit_22_expanding
    unit_23_expanding
    unit_24_expanding
    unit_25_expanding
    unit_26_expanding
    unit_27_expanding
    unit_28_expanding
    unit_28_exit_window
    unit_29_entry_window
    unit_29_straight
    unit_30_straight
    unit_31_straight
    unit_32_straight
    unit_33_straight
    unit_34_straight
    unit_35_straight
    unit_36_straight
    unit_37_straight
    unit_38_straight
    unit_39_straight
    unit_40_straight
    unit_41_straight
    unit_42_straight
    unit_43_straight
    bandwidth_chopper_1
    bandwidth_chopper_2
    unit_43_exit_window
    bandwidth_monitor
    unit_44_entry_window
    unit_44_straight
    unit_45_straight
    unit_46_straight
    unit_47_straight
    unit_48_straight
    unit_49_straight
    unit_50_straight
    unit_51_straight
    unit_52_straight
    unit_53_straight
    unit_54_straight
    unit_55_straight
    unit_56_straight
    unit_57_straight
    unit_58_straight
    unit_59_straight
    unit_60_straight
    unit_61_straight
    unit_62_straight
    unit_63_straight
    unit_64_straight
    unit_65_straight
    unit_66_straight
    unit_67_straight
    unit_68_straight
    unit_69_straight
    unit_70_straight
    unit_71_straight
    unit_72_straight
    unit_73_straight
    unit_74_straight
    unit_75_straight
    unit_76_closing
    unit_77_closing
    unit_78_closing
    unit_79_closing
    unit_80_closing
    unit_81_closing
    unit_82_closing
    unit_83_closing
    unit_84_closing
    unit_85_closing
    jaw_3
    unit_86_closing
    jaw_2
    unit_87_closing
    jaw_1
    unit_88_closing
    unit_88_exit_window
    mask
    normalization_monitor
    slit
    sample_origin""").splitlines()

    for name, ex in zip(names, expected, strict=True):
        assert name == ex

    assert len(parameters) == len(expected)


def beam_dists(comp, start):
    """Compute the distance(s) to a component's or its segments' beam position

    That is the component position plus its offset vector, if it has one.
    """
    from scipp import norm, vector
    if hasattr(comp, 'segments'):
        return [y for x in comp.segments for y in beam_dists(x, start)]
    offset = getattr(comp, 'offset', vector([0, 0, 0.], unit=comp.position.unit))
    beam_at = comp.position + offset
    return [norm(beam_at - start)]


def test_primary_create():
    from niess.bifrost.parameters import primary_parameters
    from niess.bifrost.primary import Primary
    from scipp import norm, scalar, isclose
    parameters = primary_parameters()
    primary = Primary.from_calibration(parameters)
    # ... just getting here is a test that the from_calibration mechanism works

    # each successive component should be farther away from the source:
    start = primary.source.position
    last = norm(start - start)
    for part in primary.parts()[1:]:
        for dist in beam_dists(getattr(primary, part), start):
            assert dist > last, f'Positioning error for {part}: {dist} <= {last}'
            last = dist

    start_to_end = primary.slit.position - primary.source.position
    # the sample is ~162 m from the source; but the primary spectrometer ends at
    # the slit which is ~0.5 m from the sample
    assert isclose(norm(start_to_end), scalar(161.5, unit='m'), atol=scalar(0.5, unit='m'))

    sample_at = primary.sample_origin.position - primary.source.position
    assert isclose(norm(sample_at), scalar(161.75, unit='m'), atol=scalar(0.1, unit='m'))


def test_primary_serialize_deserialize():
    from json import dumps, loads
    from niess.bifrost.parameters import primary_parameters
    from niess.bifrost.primary import Primary
    from niess.utilities import serializable
    from dataclasses import asdict
    from niess.utilities import compare
    parameters = primary_parameters()
    primary = Primary.from_calibration(parameters)
    pdict = serializable(asdict(primary))
    ddict = loads(dumps(pdict, indent=4))
    assert compare(pdict, ddict)


from dataclasses import dataclass
from scipp import Variable
from mccode_antlr.assembler import Assembler


@dataclass
class Component:
    """Any component in the instrument.

    Note
    ----
    If an inheriting class adds an 'offset' attribute to the component, the
    position reported for McStas/McXtrace/McCode has that offset added to position

    Parameters
    ----------
    name: str
        The (unique) name of the component instance
    position: Vector
        The position of the component instance in a global coordinate system. This
        may differ from the position required for, e.g., McStas (see 'offset' Note).
    orientation: Quaternion
        The orientation of the component instance in scipp quaternion form. This
        transforms the coordinate system of the component into the global coordinate
        system.
    """
    name: str
    position: Variable
    orientation: Variable

    @classmethod
    def from_calibration(cls, calibration: dict):
        name = calibration['name']
        position = calibration['position']
        orientation = calibration['orientation']
        return cls(name, position, orientation)

    def __mccode__(self) -> tuple[str, dict]:
        """Return the component type name and parameters needed to produce a McCode instance"""
        return 'Arm', {}

    def to_mccode(self, assembler: Assembler):
        from mccode_antlr.common.parameters import InstrumentParameter as InstPar
        from ..spatial import mccode_ordered_angles
        from ..mccode import ensure_runtime_parameter

        comp, pars = self.__mccode__()

        if len(pairs:=[(k, x) for k, x in pars.items() if isinstance(x, InstPar)]):
            for name, value in pairs:
                ensure_runtime_parameter(assembler, value)
                pars[name] = str(value)

        at = self.position
        if hasattr(self, 'offset'):
            at += getattr(self, 'offset')
        at = at.to(unit='m').value
        rot = mccode_ordered_angles(self.orientation)

        return assembler.component(self.name, comp, at=at, rotate=rot, parameters=pars)


from __future__ import annotations

from dataclasses import dataclass

@dataclass
class Pack:
    from mccode_antlr.assembler import Assembler
    from scipp import Variable
    from ..components import He3Tube

    tubes: tuple[He3Tube, ...]
    resistances: Variable

    @staticmethod
    def from_calibration(position: Variable, length: Variable, **params):
        from numpy import tile
        from scipp import vector, scalar, Variable, concat, ones
        from scipp.spatial import rotations
        from ..spatial import is_scipp_vector
        from ..utilities import is_type, is_scalar
        from ..components import He3Tube

        is_scipp_vector(position, 'position')
        if position.ndim != 1 or 'tube' not in position.dims:
            raise RuntimeError("Expected a 1-D list of 'tube' positions")

        ori = params.get('orient', None)
        ori = rotations(values=tile([0, 0, 0, 1], (position.sizes['tube'], 1)), dims=['tube']) if ori is None else ori

        pressure = params.get('pressure', scalar(1., unit='atm'))
        radius = params.get('radius', scalar(25.4/2, unit='mm'))
        elements = params.get('elements', 100)
        resistivity = params.get('resistivity', scalar(140., unit='Ohm/in').to(unit='Ohm/m'))

        map(lambda x: is_type(*x), ((pressure, Variable, 'pressure'), (length, Variable, 'length'),
                                    (radius, Variable, 'radius'), (elements, int, elements),
                                    (resistivity, Variable, 'resistivity')))
        pack = elements, radius, pressure

        if is_scalar(resistivity):
            resistivity = resistivity * ones(shape=(position.sizes['tube'], ), dims=['tube'])

        # Make the oriented tube axis vector(s)
        axis = ori * (length.to(unit=position.unit) * vector([0, 0, 1.]))  # may be a 0-D or 1-D tube vector array
        tube_at = position - 0.5 * axis
        tube_to = position + 0.5 * axis
        tubes = tuple(He3Tube(at, to, rho, *pack) for at, to, rho in zip(tube_at, tube_to, resistivity))

        # Define the contact resistance for the wires
        resistance = params.get('resistance', scalar(2, unit='Ohm'))
        if is_scalar(resistance):
            resistance = resistance * ones(shape=(position.sizes['tube'], 2), dims=['tube', 'end'])
        # allow overriding specific resistances ... somehow
        return Pack(tubes, resistance)

    def mcstas_parameters(self) -> dict:
        from scipp import sqrt, dot
        # length vector (from one end of each tube to the other)
        lv = [tube.to - tube.at for tube in self.tubes]
        # central vector (the position of the center, relative to defining sample position)
        cv = [(tube.to + tube.at) / 2 for tube in self.tubes]
        # average length -- they *should* all be identical, but maybe they're not?
        length = sum([sqrt(dot(x, x)).to(unit='m').value for x in lv]) / len(self.tubes)
        # average radius -- ditto, if they're not identical this is wrong and a vector in the instrument is better
        radius = sum([tube.radius.to(unit='m').value for tube in self.tubes]) / len(self.tubes)
        # The distance between the first and last tube centres plus twice the radius is the assembly width
        width = sqrt(dot(cv[-1] - cv[0], cv[-1] - cv[0])).to(unit='m').value + 2 * radius
        params = dict(
            charge_a='"event_charge_top"',
            charge_b='"event_charge_bottom"',
            detection_time='"event_time"',
            tube_index_name='"TUBE"',
            N=len(self.tubes),
            width=width,
            height=length,
            radius=radius,
            wires_in_series=0,
        )
        return params

    def to_mccode(
            self,
            assembler: Assembler,
            relative: str,
            distance: float,
            name: str,
            when: str | None = None,
            extend: str | None = None,
            add_metadata: bool = False,
            component: str | None = None,
            group: str | None = None,
            **kwargs
    ):
        if component is None:
            component = 'Detector_tubes'
        params = self.mcstas_parameters()
        #TODO Ensure that one of the provided parameters is 'first_wire_index'
        #     Another useful parameter would be 'pack_filename', to save the McStas
        #     2-D histogram directly
        params.update(kwargs)
        tubes = assembler.component(name, component, at=((0, 0, distance), relative), parameters=params)
        if when:
            tubes.WHEN(when)
        if extend:
            tubes.EXTEND(extend)
        if group:
            tubes.GROUP(group)
